/**
 * @file actuator.hpp
 * @author Nawe Robotics
 * @brief File contains IMU class definations
 */

#pragma once
#include "can_device_type.hpp"
#include "interface.hpp"

/**
 * @brief IMU Class
 *
 */
class IMU : public CANDeviceType {

    /**
     * @brief Class defination for IMU Implementation class
     *
     */
    class IMUImpl;

    /**
     * @brief Pointer to class object which stores IMU implementation
     *
     */
    IMUImpl *impl;

    /**
     * @brief CAN Interface object reference. This object is used to send/receive
     * data from CAN Bus.
     *
     */
    Interface *iface;

    /**
     * @brief Callback function which is invoked when a new CAN msg for this
     * device comes into the interface
     *
     * @param fra Reference to the received CAN Msg frame
     */
    void canMsgCallback(struct can_frame_data *fra);
public:
    /**
     * @brief Construct a new IMU object
     *
     * @param device_id Device Id for the IMU device
     * @param interface CAN Interface object for the interface to which the
     * actuator is connected
     */
    IMU(uint32_t device_id, Interface *interface);

    /**
     * @brief Set the Output Log Stream object
     *
     * @param out Output log stream object
     */
    void setOutputLogStream(std::ostream *out);

    /*Controller CAN Functions*/
    /**
     * @brief Sends Emergency Stop Message to stop actuator
     *
     * @return Return Status
     */
    ret_status_t eStop();

    /**
     * @brief Function to get Yaw data of the IMU
     * @param yaw yaw of the imu
     * @return Return Status
    */
    ret_status_t getYawData(float* yaw);

    /**
     * @brief Function to get Pitch data of the IMU
     * @param pitch pitch of the imu
     * @return Return Status
    */
    ret_status_t getPitchData(float* pitch);

    /**
     * @brief Function to get Roll data of the IMU
     * @param pitch roll of the imu
     * @return Return Status
    */
    ret_status_t getRollData(float* roll);

    /**
     * @brief Destroy the IMU object
     *
     */
    ~IMU();

    /**
     * @brief Check if the device is still connected or not
     *
     * @return true
     * @return false
     */
    bool isConnected();

    /**
     * @brief Function to flash a program onto the device
     * 
     * @param firmware_path path to firmware file
     * @return ret_status_t 
     */
    ret_status_t flash(std::string firmware_path);
};